import java.util.*;

class Pair<T, U> {
  public T a;
  public U b;

  Pair(T a, U b) {
    this.a = a;
    this.b = b;
  }
}

class Corpus {

  // All words and quick lookup 
  public ArrayList<ArrayList<Word>> yearlyCorpus;
  public Map<String, ArrayList<Pair<Integer, Integer>>> quickLookup;

  // Quick Lookup for selected words
  public Set<String> selected;
  public ArrayList<ArrayList<Word>> activePoints;
  public ArrayList<Map<String, Integer> > activeQuickLookup;

  public int startYear = 2008;
  public int start = 0;
  public int end = 200;
  public int totalNYears = 0;

  float globalWindowedMaxCount = 0;
  float globalWindowedMinCount = 1000000;

  Corpus(String yearlyCorpusFileName) {
    Table table = loadTable(yearlyCorpusFileName, "header");
    int currentYear = startYear;
    int currentIdx = 0;
    this.yearlyCorpus = new ArrayList<ArrayList<Word>>();
    this.selected = new HashSet<String>();
    this.activePoints = new ArrayList<ArrayList<Word>>();
    this.activeQuickLookup = new ArrayList<Map<String, Integer>>();
    this.quickLookup = new HashMap<String, ArrayList<Pair<Integer, Integer>>>();

    this.yearlyCorpus.add(new ArrayList<Word>());
    for (int i = 0; i < table.getRowCount(); i++) {
      // Start of a new year, create a new list
      if (table.getInt(i, 0) != currentYear) {
        currentYear = table.getInt(i, 0);
        Collections.sort(this.yearlyCorpus.get(currentIdx), Collections.reverseOrder());
        currentIdx += 1;
        this.yearlyCorpus.add(new ArrayList<Word>());
      }

      // Add a new word to the list
      int year = table.getInt(i, 0);
      String word = table.getString(i, 1);
      float count = log(table.getFloat(i, 2)) / log(10);
      float x = table.getFloat(i, 3);
      float y = table.getFloat(i, 4);
      float z = table.getFloat(i, 5);

      // Filter out counts 
      if (count > log(COUNT_FILTER)/log(10)) {
        this.yearlyCorpus.get(currentIdx).add(new Word(year, word, count, x, y, z, 255));
      }
    }
    totalNYears = currentIdx;

    // Handle last year
    Collections.sort(this.yearlyCorpus.get(currentIdx), Collections.reverseOrder());

    for (int i = 0; i < yearlyCorpus.size(); i++) {
      for (int j = 0; j < yearlyCorpus.get(i).size(); j++) {
        String word = yearlyCorpus.get(i).get(j).word;
        if (!quickLookup.containsKey(word)) {
          quickLookup.put(word, new ArrayList<Pair<Integer, Integer>>());
        }
        quickLookup.get(word).add(new Pair(i, j));
      }
    }

    this.setActivePoints(0, 100);
  }

  public void setActivePoints(int start, int end) {
    this.activePoints.clear();
    this.activeQuickLookup.clear();
    this.start = start;
    this.end = end;

    this.globalWindowedMaxCount = 0;

    // Rescale windowed max for color mapping
    for (int i = 0; i < this.yearlyCorpus.size(); i++) {
      for (int j = start; j < end; j++) {
        if (this.yearlyCorpus.get(i).get(j).count > this.globalWindowedMaxCount) { 
          this.globalWindowedMaxCount = this.yearlyCorpus.get(i).get(j).count;
        };
        if (this.yearlyCorpus.get(i).get(j).count < this.globalWindowedMinCount) { 
          this. globalWindowedMinCount = this.yearlyCorpus.get(i).get(j).count;
        };
      }
    }

    // Set new set of active points
    for (int i = 0; i < this.yearlyCorpus.size(); i++) {
      this.activePoints.add(rescaleYears(this.yearlyCorpus.get(i), start, end, this.globalWindowedMinCount, this.globalWindowedMaxCount));
      Map<String, Integer> ql = new HashMap<String, Integer>();
      for (int j = 0; j < this.activePoints.get(i).size(); j++) {
        ql.put(this.activePoints.get(i).get(j).word, j);
      }
      this.activeQuickLookup.add(ql);
    }
  }

  void setSelectedWords(String selectedWords) {
    selectedWords = selectedWords.toLowerCase();
    this.selected.clear();
    if (selectedWords.toLowerCase().equals("--all")) {
      for (ArrayList<Word> words : activePoints) {
        for (Word word : words) {
          selected.add(word.word);
        }
      }
    } else {
      String[] sWordArr = selectedWords.split(",");
      for (String word : sWordArr) {            
        if (word.trim().length() > 4) {
          for (Map<String, Integer> ql : this.activeQuickLookup) {
            if (ql.containsKey(word.trim())) {
              selected.add(word.trim());
              break;
            }
          }
        }
      }
    }
  }
}

public ArrayList<Word> rescaleYears(ArrayList<Word> words, int start, int end, float globalLocalMinCount, float globalLocalMaxCount) {

  float min_count = 100000;
  float max_count = -100000;
  float min_x = 100000;
  float max_x = -100000;
  float min_y = 100000;
  float max_y = -100000;
  float min_z = 100000;
  float max_z = -100000;

  if (end == -1 || end > words.size()) {
    end = words.size();
  }

  ArrayList<Word> scaledWords = new ArrayList<Word>();

  for (int j = start; j < end; j++) {
    float count = words.get(j).count;
    float x = words.get(j).x;
    float y = words.get(j).y;
    float z = words.get(j).z;

    if (x < min_x) { 
      min_x = x;
    }
    if (y < min_y) { 
      min_y = y;
    }
    if (z < min_z) { 
      min_z = z;
    }
    if (count < min_count) { 
      min_count = count;
    }

    if (x > max_x) { 
      max_x = x;
    }
    if (y > max_y) { 
      max_y = y;
    }
    if (z > max_z) { 
      max_z = z;
    }
    if (count > max_count) { 
      max_count = count;
    }
  }

  for (int j = start; j < end; j++) {

    // x and z are mapped as raw coordinates
    float x = map(words.get(j).x, min_x, max_x, -MAP_X, MAP_X);
    float z = map(words.get(j).z, min_z, max_z, -MAP_Z, MAP_Z);  

    // y is mapped based on popularity
    float y = map(j, start, end, -MAP_Y, MAP_Y); 

    // count is mapped to a font-size
    float count  = map(words.get(j).count, min_count, max_count, 16, MAP_COUNT_RADIUS);

    // absolute count is also mapped to a color value
    float cv = map(words.get(j).count, min_count, max_count, 0, 1);
    //float cv = round(map(j, start, end, 0, 1));

    color c = lerpColor(MIN_COLOR, MAX_COLOR, cv);

    scaledWords.add(new Word(words.get(j).year, words.get(j).word, count, x, y, z, c));
  }

  return scaledWords;
}
